@@ -76,7 +76,7 @@ struct _LmConnection {
 	guint         io_watch_err;
 	guint         io_watch_hup;
 	guint         fd;
-    guint         io_watch_connect;
+	guint         io_watch_connect;
 	
 	guint         open_id;
 	LmCallback   *open_cb;
@@ -96,6 +96,9 @@ struct _LmConnection {
 	guint         keep_alive_rate;
 	guint         keep_alive_id;
 
+	guint         io_watch_out;
+	GString      *out_buf;
+
 	gint          ref_count;
 };
 
@@ -118,6 +121,10 @@ static gboolean connection_do_open           (LmConnection         *connection,
 					      GError              **error);
 
 static void     connection_do_close           (LmConnection        *connection);
+static gint     connection_do_write          (LmConnection         *connection,
+					      const gchar          *buf,
+					      gint                  len);
+
 static gboolean connection_in_event          (GIOChannel   *source,
 					      GIOCondition  condition,
 					      LmConnection *connection);
@@ -127,7 +134,7 @@ static gboolean connection_error_event       (GIOChannel   *source,
 static gboolean connection_hup_event         (GIOChannel   *source,
 					      GIOCondition  condition,
 					      LmConnection *connection);
-static gboolean connection_send              (LmConnection             *connection,
+static gboolean connection_send              (LmConnection         *connection,
 					      const gchar          *str,
 					      gint                  len,
 					      GError               **error);
@@ -148,8 +155,8 @@ static LmHandlerResult connection_auth_reply (LmMessageHandler    *handler,
 					      LmMessage           *m,
 					      gpointer             user_data);
 
-static void     connection_stream_received   (LmConnection             *connection, 
-					      LmMessage                *m);
+static void     connection_stream_received   (LmConnection        *connection, 
+					      LmMessage           *m);
 
 static gint     connection_handler_compare_func (HandlerData  *a,
 						 HandlerData  *b);
@@ -172,6 +179,15 @@ static guint    connection_add_watch            (LmConnection  *connection,
 static gboolean connection_send_keep_alive      (LmConnection  *connection);
 static void     connection_start_keep_alive     (LmConnection  *connection);
 static void     connection_stop_keep_alive      (LmConnection  *connection);
+static gboolean connection_buffered_write_cb    (GIOChannel    *source, 
+						 GIOCondition   condition,
+						 LmConnection  *connection);
+static gboolean connection_output_is_buffered   (LmConnection  *connection,
+						 const gchar   *buffer,
+						 gint           len);
+static void     connection_setup_output_buffer  (LmConnection  *connection,
+						 const gchar   *buffer,
+						 gint           len);
 
 static GSourceFuncs incoming_funcs = {
 	connection_incoming_prepare,
@@ -217,7 +233,16 @@ connection_free (LmConnection *connection)
 	}
 
 	g_queue_free (connection->incoming_messages);
-	g_free (connection);
+        
+        if (connection->context) {
+                g_main_context_unref (connection->context);
+        }
+
+	if (connection->out_buf) {
+		g_string_free (connection->out_buf, TRUE);
+	}
+        
+        g_free (connection);
 }
 
 static void
@@ -360,7 +385,9 @@ _lm_connection_succeeded (LmConnectData *connect_data)
 				      G_IO_HUP,
 				      (GIOFunc) connection_hup_event,
 				      connection);
-	
+
+	/* FIXME: Set up according to XMPP 1.0 specification */
+	/*        StartTLS and the like */
 	if (!connection_send (connection, 
 			      "<?xml version='1.0' encoding='UTF-8'?>", -1,
 			      NULL)) {
@@ -419,13 +446,14 @@ _lm_connection_failed_with_error (LmConnectData *connect_data, int error)
 	}
 	
 	if (connect_data->current_addr == NULL) {
+		connection_do_close (connection);
 		if (connection->open_cb && connection->open_cb->func) {
 			LmCallback *cb = connection->open_cb;
 			
 			(* ((LmResultFunction) cb->func)) (connection, FALSE,
 							   cb->user_data);
 		}
-
+		
 		freeaddrinfo (connect_data->resolved_addrs);
 		g_free (connect_data);
 	} else {
@@ -447,7 +475,7 @@ connection_connect_cb (GIOChannel   *source,
 {
 	LmConnectData *connect_data;
 	int            error;
-	int            len  = sizeof(error);
+	guint          len  = sizeof(error);
 
 	connect_data = (LmConnectData *) data;
 	
@@ -521,14 +549,14 @@ connection_do_connect (LmConnectData *connect_data)
 	
 	connect_data->io_channel = g_io_channel_unix_new (fd);
 	if (connection->proxy) {
-        connection->io_watch_connect =
+		connection->io_watch_connect =
 		connection_add_watch (connection,
 				      connect_data->io_channel,
 				      G_IO_OUT|G_IO_ERR,
 				      (GIOFunc) _lm_proxy_connect_cb, 
 				      connect_data);
 	} else {
-        connection->io_watch_connect =
+		connection->io_watch_connect =
 		connection_add_watch (connection,
 				      connect_data->io_channel,
 				      G_IO_OUT|G_IO_ERR,
@@ -556,6 +584,7 @@ connection_add_watch (LmConnection *connection,
 	g_source_set_callback (source, (GSourceFunc)func, user_data, NULL);
                                                                                 
 	id = g_source_attach (source, connection->context);
+
 	g_source_unref (source);
   
 	return id;
@@ -596,6 +625,75 @@ connection_stop_keep_alive (LmConnection *connection)
 	connection->keep_alive_id = 0;
 }
 
+static gboolean
+connection_buffered_write_cb (GIOChannel   *source, 
+			      GIOCondition  condition,
+			      LmConnection *connection)
+{
+	gint     b_written;
+	GString *out_buf;
+	/* FIXME: Do the writing */
+
+	out_buf = connection->out_buf;
+	if (!out_buf) {
+		/* Should not be possible */
+		return FALSE;
+	}
+
+	b_written = connection_do_write (connection, out_buf->str, out_buf->len);
+
+	if (b_written < 0) {
+		connection_error_event (connection->io_channel, 
+					G_IO_HUP,
+					connection);
+		return FALSE;
+	}
+
+	g_string_erase (out_buf, 0, (gsize) b_written);
+	if (out_buf->len == 0) {
+		lm_verbose ("Output buffer is empty, going back to normal output\n");
+		g_source_destroy (g_main_context_find_source_by_id (
+			connection->context, connection->io_watch_out));
+		connection->io_watch_out = 0;
+		g_string_free (out_buf, TRUE);
+		connection->out_buf = NULL;
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static gboolean
+connection_output_is_buffered (LmConnection *connection,
+			       const gchar  *buffer,
+			       gint          len)
+{
+	if (connection->out_buf) {
+		lm_verbose ("Appending %d bytes to output buffer\n", len);
+		g_string_append_len (connection->out_buf, buffer, len);
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static void
+connection_setup_output_buffer (LmConnection *connection,
+				const gchar  *buffer,
+				gint          len)
+{
+	lm_verbose ("OUTPUT BUFFER ENABLED\n");
+
+	connection->out_buf = g_string_new_len (buffer, len);
+
+	connection->io_watch_out =
+		connection_add_watch (connection,
+				      connection->io_channel,
+				      G_IO_OUT,
+				      (GIOFunc) connection_buffered_write_cb,
+				      connection);
+}
+
 /* Returns directly */
 /* Setups all data needed to start the connection attempts */
 static gboolean
@@ -608,7 +706,7 @@ connection_do_open (LmConnection *connection, GError **error)
 	if (lm_connection_is_open (connection)) {
 		g_set_error (error,
 			     LM_ERROR,
-			     LM_ERROR_CONNECTION_NOT_OPEN,
+			     LM_ERROR_CONNECTION_OPEN,
 			     "Connection is already open, call lm_connection_close() first");
 		return FALSE;
 	}
@@ -616,7 +714,7 @@ connection_do_open (LmConnection *connection, GError **error)
 	if (!connection->server) {
 		g_set_error (error,
 			     LM_ERROR,
-			     LM_ERROR_CONNECTION_OPEN,
+			     LM_ERROR_CONNECTION_FAILED,
 			     "You need to set the server hostname in the call to lm_connection_new()");
 		return FALSE;
 	}
@@ -648,7 +746,7 @@ connection_do_open (LmConnection *connection, GError **error)
 		if (getaddrinfo (proxy_server, NULL, &req, &ans) != 0) {
 			g_set_error (error,
 				     LM_ERROR,                 
-				     LM_ERROR_CONNECTION_OPEN,   
+				     LM_ERROR_CONNECTION_FAILED,   
 				     "getaddrinfo() failed");
 			return FALSE;
 		}
@@ -661,7 +759,7 @@ connection_do_open (LmConnection *connection, GError **error)
 				 NULL, &req, &ans) != 0) {
 			g_set_error (error,
 				     LM_ERROR,                 
-				     LM_ERROR_CONNECTION_OPEN,   
+				     LM_ERROR_CONNECTION_FAILED,   
 				     "getaddrinfo() failed");
 			return FALSE;
 		}
@@ -697,6 +795,12 @@ connection_do_close (LmConnection *connection)
 		g_source_destroy (g_main_context_find_source_by_id (
 			connection->context, connection->io_watch_hup));
 
+		if (connection->io_watch_out != 0) {
+			g_source_destroy (g_main_context_find_source_by_id (
+				connection->context, connection->io_watch_out));
+			connection->io_watch_out = 0;
+		}
+
 		if (connection->io_watch_connect != 0) {
 			g_source_destroy (g_main_context_find_source_by_id(connection->context,
 									   connection->io_watch_connect));
@@ -708,22 +812,51 @@ connection_do_close (LmConnection *connection)
 
 		connection->fd = -1;
 	}
-
 	
 	g_source_destroy (connection->incoming_source);
 	g_source_unref (connection->incoming_source);
 
 	if (!lm_connection_is_open (connection)) {
+		/* lm_connection_is_open is FALSE for state OPENING as well */
+		connection->state = LM_CONNECTION_STATE_CLOSED;
 		return;
 	}
 	
 	connection->state = LM_CONNECTION_STATE_CLOSED;
-
 	if (connection->ssl) {
 		_lm_ssl_close (connection->ssl);
 	}
 }
 
+static gint
+connection_do_write (LmConnection *connection,
+		     const gchar  *buf,
+		     gint          len)
+{
+	gint b_written;
+
+	if (connection->ssl) {
+		b_written = _lm_ssl_send (connection->ssl, buf, len);
+	} else {
+		GIOStatus io_status = G_IO_STATUS_AGAIN;
+		gsize     bytes_written;
+
+		while (io_status == G_IO_STATUS_AGAIN) {
+			io_status = g_io_channel_write_chars (connection->io_channel, 
+							      buf, len, 
+							      &bytes_written,
+							      NULL);
+		}
+
+		b_written = bytes_written;
+
+		if (io_status != G_IO_STATUS_NORMAL) {
+			b_written = -1;
+		}
+	}
+
+	return b_written;
+}
 
 static gboolean
 connection_in_event (GIOChannel   *source,
@@ -773,14 +906,14 @@ connection_in_event (GIOChannel   *source,
 
 	buf[bytes_read] = '\0';
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "\nRECV [%d]:\n", 
-	       bytes_read);
+	       (int)bytes_read);
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
 	       "-----------------------------------\n");
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "'%s'\n", buf);
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
 	       "-----------------------------------\n");
 
-	lm_verbose ("Read: %d chars\n", bytes_read);
+	lm_verbose ("Read: %d chars\n", (int)bytes_read);
 
 	lm_parser_parse (connection->parser, buf);
 	
@@ -827,7 +960,7 @@ connection_send (LmConnection  *connection,
 		 gint           len, 
 		 GError       **error)
 {
-	gsize             bytes_written;
+	gint b_written;
 	
 	if (connection->state < LM_CONNECTION_STATE_OPENING) {
 		g_set_error (error,
@@ -848,16 +981,27 @@ connection_send (LmConnection  *connection,
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
 	       "-----------------------------------\n");
 
-	if (connection->ssl) {
-		if (!_lm_ssl_send (connection->ssl, str, len)) {
-			
-			connection_error_event (connection->io_channel, 
-						G_IO_HUP,
-						connection);
-		}
-	} else {
-		g_io_channel_write_chars (connection->io_channel, str, len, 
-					  &bytes_written, NULL);
+	/* Check to see if there already is an output buffer, if so, add to the
+	   buffer and return */
+
+	if (connection_output_is_buffered (connection, str, len)) {
+		return TRUE;
+	}
+
+	b_written = connection_do_write (connection, str, len);
+
+
+	if (b_written < 0) {
+		connection_error_event (connection->io_channel, 
+					G_IO_HUP,
+					connection);
+		return FALSE;
+	}
+
+	if (b_written < len) {
+		connection_setup_output_buffer (connection, 
+						str + b_written, 
+						len - b_written);
 	}
 
 	return TRUE;
@@ -1172,6 +1316,7 @@ lm_connection_new (const gchar *server)
 	connection->state             = LM_CONNECTION_STATE_CLOSED;
 	connection->keep_alive_id     = 0;
 	connection->keep_alive_rate   = 0;
+	connection->out_buf           = NULL;
 	
 	connection->id_handlers = g_hash_table_new_full (g_str_hash, 
 							 g_str_equal,
@@ -1209,7 +1354,7 @@ lm_connection_new_with_context (const gchar *server, GMainContext *context)
 	connection = lm_connection_new (server);
 	connection->context = context;
 
-	g_main_context_ref (connection->context);
+        g_main_context_ref (connection->context);
 
 	return connection;
 }
@@ -1278,6 +1423,12 @@ lm_connection_open_and_block (LmConnection *connection, GError **error)
 		return TRUE;
 	}
 
+	/* Need to set the error here: LM-15 */
+	g_set_error (error,
+		     LM_ERROR,
+		     LM_ERROR_CONNECTION_FAILED,
+		     "Opening the connection failed");
+
 	return FALSE;
 }
 
@@ -1376,6 +1527,8 @@ lm_connection_authenticate (LmConnection      *connection,
 		return FALSE;
 	}
 
+	/* FIXME: Do SASL authentication here (if XMPP 1.0 is used) */
+
 	connection->state = LM_CONNECTION_STATE_AUTHENTICATING;
 	
 	connection->auth_cb = _lm_utils_new_callback (function, 
@@ -1436,11 +1589,14 @@ lm_connection_authenticate_and_block (LmConnection  *connection,
 		return FALSE;
 	}
 
+	connection->state = LM_CONNECTION_STATE_AUTHENTICATING;
+
 	m = connection_create_auth_req_msg (username);
 	result = lm_connection_send_with_reply_and_block (connection, m, error);
 	lm_message_unref (m);
 
 	if (!result) {
+		connection->state = LM_CONNECTION_STATE_OPEN;
 		return FALSE;
 	}
 
@@ -1453,6 +1609,7 @@ lm_connection_authenticate_and_block (LmConnection  *connection,
 
 	result = lm_connection_send_with_reply_and_block (connection, m, error);
 	if (!result) {
+		connection->state = LM_CONNECTION_STATE_OPEN;
 		return FALSE;
 	}
 
@@ -1461,9 +1618,11 @@ lm_connection_authenticate_and_block (LmConnection  *connection,
 	
 	switch (type) {
 	case LM_MESSAGE_SUB_TYPE_RESULT:
+		connection->state = LM_CONNECTION_STATE_AUTHENTICATED;
 		return TRUE;
 		break;
 	case LM_MESSAGE_SUB_TYPE_ERROR:
+		connection->state = LM_CONNECTION_STATE_OPEN;
 		g_set_error (error,
 			     LM_ERROR,
 			     LM_ERROR_AUTH_FAILED,
@@ -1676,6 +1835,7 @@ void
 lm_connection_set_ssl (LmConnection *connection, LmSSL *ssl)
 {
 	g_return_if_fail (connection != NULL);
+	g_return_if_fail (lm_ssl_is_supported () == TRUE);
 
 	if (connection->ssl) {
 		lm_ssl_unref (connection->ssl);
@@ -1823,6 +1983,15 @@ lm_connection_send_with_reply_and_block (LmConnection  *connection,
 	g_return_val_if_fail (connection != NULL, NULL);
 	g_return_val_if_fail (message != NULL, NULL);
 
+	if (connection->state < LM_CONNECTION_STATE_OPENING) {
+		g_set_error (error,
+			     LM_ERROR,
+			     LM_ERROR_CONNECTION_NOT_OPEN,
+			     "Connection is not open, call lm_connection_open() first");
+		return FALSE;
+	}
+
+
 	if (lm_message_node_get_attribute (message->node, "id")) {
 		id = g_strdup (lm_message_node_get_attribute (message->node, 
 							      "id"));
